#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Wed May 29 09:29:13 2019

@author: nvthaomy

Simulating Gaussian polymer/fluid in NVT/NPT ensemble
With or without the external potential

"""
import sim
import mdtraj as md
from simtk import openmm, unit
from simtk.unit import *
from simtk.openmm import app
import numpy as np
import simtk.openmm.app.topology as topology

#========================================
###DEFINE ENERGY, LENGTH & MASS SCALES###
#========================================
epsilon  = 1 * kilojoules_per_mole
sigma = 0.1 * nanometer
mass = 12 * amu
N_av = 6.022140857*10**23 /mole
kb = 1.380649*10**(-23)*joules/kelvin* N_av #joules/kelvin/mol
e0 = 8.85e-12 *coulomb*(volt*meter)**-1 #permittivity of free space
#====================================
#1) SYSTEM DIMENSIONLESS PARAMETERS###
#====================================

muPAs = np.array([0] )
muPBs = np.array([-8, -4, 0, 4, 8])
#muPAs = np.array([-8,-7,-6,-5,-4,-3,-2,-1,0,1,2,3,4,5,6,7,8]) #[-8,8]
#muPBs = np.array([-8,-7,-6,-5,-4,-3,-2,-1,0,1,2,3,4,5,6,7,8]) #[-8,8]

#Molecules:
nA = 1
NA = 60
nB = 1
NB = 60
N = nA*NA + nB*NB
charge = 0.0 * elementary_charge
reduced_density = N/(100**3)
qAs = [0,-1] * int(nA*NA/2) #initial charges
qBs = [0,+1] * int(nB*NB/2)
charges = qAs + qBs
deprot = -1
neutral = 0
prot = +1

#Bonded potential 3/2*kbT*(r-r0)^2/b0^2
b0 = 1.1 #reduced bond length
r0 = 0
#k = 3 kbT/b0^2

#Pair potential: u = 4*epsilonP((sigmaP/r)^12 - (sigmaP/r)^6)
reduced_nonbondedCutoff = 2**(1/6)

#Electrostatic parameters
reduced_lB = 1
reduced_lD = 1/0.1 #[1/6,1/0.1]
reduced_eleCutoff = 49.9

#======================
#2) Integration Options
#======================
stride = 200
pHMC = 1/20 #probability of performing hybridMC move
useNVT = True
useNPT = False
reduced_timestep = 0.005
reduced_temp = 1
reduced_Tdamp = 100*reduced_timestep #time units
reduced_pressure = 1
reduced_Pdamp = 25*reduced_timestep #time units
#=====================
#3) Simulation Options
#=====================
sweeps =  int(2e5)
steps = 1e3
equilibrationSteps = 1e5 #1e5 #1e7
#platform = openmm.Platform.getPlatformByName('CPU')
#platformProperties = {'Precision': 'mixed'}
#platform = openmm.Platform.getPlatformByName('CUDA')

traj = 'trajectoryMD.dcd'
dcdReporter = md.reporters.DCDReporter(traj,100)
#pdbReporter = app.pdbreporter.PDBReporter('trajectory.pdb',stride)
dataReporter = app.statedatareporter.StateDataReporter('log.txt', 500, totalSteps=steps,
    step=True, speed=True, progress=True, potentialEnergy=True, kineticEnergy=True, 
    totalEnergy=True, temperature=True, volume=True, density=True, remainingTime=True,separator='\t')

dcdReporter_warmup = md.reporters.DCDReporter('trajectoryMD_warmup.dcd',500)
dataReporter_warmup= app.statedatareporter.StateDataReporter('log_warmup.txt', 2000, totalSteps=equilibrationSteps,
    step=True, speed=True, progress=True, potentialEnergy=True, kineticEnergy=True,
    totalEnergy=True, temperature=True, volume=True, density=True, remainingTime=True,separator='\t')

nonbondedMethod = openmm.CustomNonbondedForce.CutoffPeriodic
ewaldErrorTolerance = 0.0001
constraints = None
#rigidWater = True
#constraintTolerance = 0.000001

#=============================
###Converting to real units###
#=============================
dt = reduced_timestep* (mass*sigma*sigma/epsilon)**(1/2)
temperature = reduced_temp * epsilon/kb
friction = 1/(reduced_Tdamp) * (epsilon/(mass*sigma*sigma))**(1/2)
pressure = reduced_pressure * epsilon/(sigma**3) * N_av**-1
barostatInterval = int(reduced_Pdamp/reduced_timestep)

lB = reduced_lB * sigma 
lD = reduced_lD * sigma
muPAs *= kb*temperature
muPBs *= kb*temperature 
#converting to harmonic bond parameter u = 1/2 * k * (x-l)^2
k = 3*kb*temperature/(b0*sigma)**2
r0 = r0*sigma

nonbondedCutoff = reduced_nonbondedCutoff*sigma
eleCutoff = reduced_eleCutoff*sigma

print('Bjerrum length: %2.2f nm'%(lB.value_in_unit(nanometer)))
print('Debye length:   %3.2f nm'%(lD.value_in_unit(nanometer)))
print ('temperature:   %3.2f K'%(temperature.value_in_unit(kelvin)))
print ('pressure:      %3.2f atm'%(pressure.value_in_unit(atmosphere)))
print ('friction:      %3.3f 1/picosecond'%(friction.value_in_unit(picosecond**-1)))
print ('barostat interval: {}'.format(barostatInterval))
print ('nonbondedCutoff: %3.3f nm'%(nonbondedCutoff.value_in_unit(nanometer)))
print ('eleCutoff:       %3.3f nm'%(eleCutoff.value_in_unit(nanometer)))
#========================================
# Create a system and add particles to it
#========================================
system = openmm.System()
# Set the periodic box vectors:
number_density = reduced_density / sigma**3
volume = N * (number_density ** -1)
box_edge = [volume ** (1. / 3.)] * 3
box_vectors = np.diag([edge/angstrom for edge in box_edge]) * angstroms
system.setDefaultPeriodicBoxVectors(*box_vectors)
print ('Box length:  %3.3f nm'%(box_edge[0].value_in_unit(nanometer)))

# Particles are added one at a time
# Their indices in the System will correspond with their indices in the Force objects we will add later
print ("Adding {} polyacid atoms into system".format(nA*NA))
for index in range(nA*NA):
    system.addParticle(mass)
print ("Adding {} polybase atoms into system".format(nB*NB))
for index in range(nB*NB):
    system.addParticle(mass)
print("Total number of paricles in system: {}".format(system.getNumParticles()))
#==================
##CREATE TOPOLOGY##
#==================
#Topology consists of a set of Chains 
#Each Chain contains a set of Residues, 
#and each Residue contains a set of Atoms.
nmols = [nA, nB]
residues = [["Acid"], ["Base"]]
PolyacidAtomList = NA*['A']
PolybaseAtomList = NB*['B']
atomList = {"Acid":PolyacidAtomList,"Base":PolybaseAtomList} #list of atoms in each residue
elements = {"A":app.element.Element(200, 'Acid', 'Ap', mass),
            "B":app.element.Element(201, 'Base', 'Bp', mass)}
def makeTop(nA,nB):
    nmols = [nA, nB]
    top = topology.Topology()
    for ispec in range(len(nmols)): #loop over each species
        for imol in range(nmols[ispec]): #loop over each molecule within species
            chain = top.addChain() #create chain for each molecule
            for ires in range( len(residues[ispec]) ): #polymer and solvent each has one residue
                resname = residues[ispec][ires]
                res = top.addResidue( resname, chain)
                atoms = atomList[resname]
                for atomInd,atomName in enumerate(atoms):
                    el = elements[atomName]
                    if atomInd > 0:
                        previousAtom = atom
                    atom = top.addAtom( atomName, el, res )
                    if atomInd > 0:
                        top.addBond(previousAtom,atom)
    return top
print ("Creating topology")
top = makeTop(nA,nB)

#=========================
#create HarmonicBondForce#
#=========================
bondedForce = openmm.HarmonicBondForce()
atomIndThusFar = 0 
for imol in range(nA): #loop over all polymer chain
    counter = 0
    for atomInd in range(atomIndThusFar,atomIndThusFar+NA-1):
        bondedForce.addBond(atomInd,atomInd+1,r0,k)
        counter += 1
    atomIndThusFar += counter + 1 #skip the last atom in polymer chain
for imol in range(nB): #loop over all polymer chain
    counter = 0
    for atomInd in range(atomIndThusFar,atomIndThusFar+NB-1):
        bondedForce.addBond(atomInd,atomInd+1,r0,k)
        counter += 1
    atomIndThusFar += counter + 1 #skip the last atom in polymer chain
system.addForce(bondedForce)

#=============================
#create custom nonbonded force
#=============================
#LJ = 4*epsilonP*((sigmaP/r)^12 - (sigmaP/r)^6) 
Polyacid = set()
Polybase = set()
for atom in top.atoms():
	if atom.residue.name in ['Acid']:
		Polyacid.add(atom.index)
	elif atom.residue.name in ['Base']:
		Polybase.add(atom.index)
all_atoms = Polyacid.union(Polybase)

LJshift = 4*epsilon*((sigma/nonbondedCutoff)**12-(sigma/nonbondedCutoff)**6)
LJ = openmm.CustomNonbondedForce('4*epsilon*((sigma/r)^12 - (sigma/r)^6)-shift')
LJ.setNonbondedMethod(nonbondedMethod)
LJ.addGlobalParameter('shift',LJshift)
LJ.addGlobalParameter('sigma',sigma)
LJ.addGlobalParameter('epsilon',epsilon)
#LJ.addInteractionGroup(Polymer,Polymer)
for i in range(system.getNumParticles()):
	LJ.addParticle()
LJ.setCutoffDistance(nonbondedCutoff)
system.addForce(LJ)
print ("Number of particles in LJ force:{}".format(LJ.getNumParticles()))
print('lj force group: {}'.format(LJ.getForceGroup()))
#Screened Coulomb = 'kbT*lB/r*exp(-r/lD)'
Coulomb = openmm.CustomNonbondedForce('kbT*charge1*charge2*lB/r*exp(-r/lD)')
Coulomb.setForceGroup(1)
Coulomb.setNonbondedMethod(nonbondedMethod)
Coulomb.addPerParticleParameter('charge')
Coulomb.addGlobalParameter('kbT',kb*temperature)
Coulomb.addGlobalParameter('lB',lB)
Coulomb.addGlobalParameter('lD',lD)
#Coulomb.addInteractionGroup(Polymer,Polymer)
for i in range(len(Polyacid)):
        Coulomb.addParticle([qAs[i]])
for i in range(len(Polybase)):
        Coulomb.addParticle([qBs[i]])
Coulomb.setCutoffDistance(eleCutoff)
system.addForce(Coulomb)
print ("Number of particles in Coulomb force:{}".format(Coulomb.getNumParticles()))
print('coulomb force group: {}'.format(Coulomb.getForceGroup()))
#===========================
## Prepare the Simulation ##
#===========================
if useNPT:
    system.addForce(openmm.MonteCarloBarostat(pressure, temperature, barostatInterval))
integrator = openmm.LangevinIntegrator(temperature, friction, dt)
simulation = app.Simulation(top,system, integrator)
#simulation = app.Simulation(top,system, integrator, platform,platformProperties)
positions = [box_edge[0]/angstrom,box_edge[1]/angstrom,box_edge[2]/angstrom]*angstrom * np.random.rand(N,3)
simulation.context.setPositions(positions)

#Restart and Check point:
#to load a saved state: simulation.loadState('output.xml')
simulation.reporters.append(app.checkpointreporter.CheckpointReporter('checkpnt.chk', 1000))

initialpdb = 'trajectory_init.pdb'
app.PDBFile.writeModel(simulation.topology, positions, open(initialpdb,'w'))
context = simulation.context

#=========================
#Set up wall force report
#=========================
class ForceReporter(object):
        def __init__(self, file, reportInterval,sweep,charges):
                self._out = open(file, 'a')
                self._fileout = file
                self._out.close()
                self._reportInterval = reportInterval
                self._hasInitialized = True
                self._sweep = sweep
                self._charges = charges
        def __del__(self):
                pass
                #self._out.close()                                            
        def describeNextReport(self, simulation):
                steps = self._reportInterval - simulation.currentStep%self._reportInterval
                return (steps, True, False, False, False, None) #(timesteps until next report, need positions, need velocities, need forces, need energies, wrapped positions)                                                             
        def report(self, simulation, state):
            self._out = open(self._fileout, 'a')
            if not self._hasInitialized: #writing header
                headers = "#sweep charges\n"
                self._out.write(headers)
                self._hasInitialized = True
            self._out.write('{} {}\n'.format(self._sweep,self._charges))
            self._out.close()

#===========================
# Minimize and Equilibrate
#===========================
if useNVT:
	ensemble = "NVT"
else:
	ensemble = "NPT"

print ("Initialize {} simulation".format(ensemble))
print ("Running energy minimization")
simulation.minimizeEnergy(maxIterations=1000)
simulation.context.setVelocitiesToTemperature(temperature*3)
print ("Running equilibration")
simulation.reporters.append(dataReporter_warmup)
simulation.reporters.append(dcdReporter_warmup)
simulation.step(equilibrationSteps)
simulation.saveState('output_warmup.xml')
state  = simulation.context.getState(getPositions=True, getVelocities=True, getForces=True, getEnergy=True, getParameters=True, getParameterDerivatives=True, enforcePeriodicBox=True)
simulation.reporters = simulation.reporters[:-2]
for muPA in muPAs:
    for muPB in muPBs:
        print ('muPA muPB: %2.1f %2.1f'%(muPA/(kb*temperature),muPB/(kb*temperature)))
        statef,checkf,finalstate,charges,simulation = sim.MC(simulation, state, system, charges, muPA, muPB, temperature, kb, sweeps, steps, stride, pHMC, Polybase, Polyacid)
        state = finalstate
        
